@@ -1,6 +1,6 @@
 /*
  * python-pctrl -- python interface to the prctl function
- * (c)2010-2018 Dennis Kaarsemaker <dennis@kaarsemaker.net>
+ * (c)2010-2020 Dennis Kaarsemaker <dennis@kaarsemaker.net>
  * See COPYING for licensing details
  */
 
@@ -50,6 +50,7 @@ prctl_prctl(PyObject *self, PyObject *args)
 {
     long option = 0;
     long arg = 0;
+    long arg2 = 0;
     char *argstr = NULL;
     char name[17] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
     int result;
@@ -59,7 +60,7 @@ prctl_prctl(PyObject *self, PyObject *args)
      * prctl possibilities. int+string is required for (and only accepted for)
      * PR_SET_NAME
      */
-    if(!PyArg_ParseTuple(args, "l|l", &option, &arg)) {
+    if(!PyArg_ParseTuple(args, "l|ll", &option, &arg, &arg2)) {
         if(!PyArg_ParseTuple(args, "ls", &option, &argstr)) {
             return NULL;
         }
@@ -88,6 +89,9 @@ prctl_prctl(PyObject *self, PyObject *args)
             break;
 #endif
         case(PR_SET_DUMPABLE):
+#ifdef PR_SET_IO_FLUSHER
+        case(PR_SET_IO_FLUSHER):
+#endif
         case(PR_SET_KEEPCAPS):
             /* Only 0 and 1 are allowed */
             arg = arg ? 1 : 0;
@@ -113,6 +117,14 @@ prctl_prctl(PyObject *self, PyObject *args)
                 return NULL;
             }
             break;
+#ifdef PR_SET_FP_MODE
+        case(PR_SET_FP_MODE):
+            if(arg != PR_FP_MODE_FR && arg != PR_FP_MODE_FRE) {
+                PyErr_SetString(PyExc_ValueError, "Unknown floating-point mode");
+                return NULL;
+            }
+            break;
+#endif
 #ifdef PR_MCE_KILL
         case(PR_SET_MCE_KILL):
             if(arg != PR_MCE_KILL_DEFAULT && arg != PR_MCE_KILL_EARLY && arg != PR_MCE_KILL_LATE) {
@@ -127,6 +139,14 @@ prctl_prctl(PyObject *self, PyObject *args)
             }
             strncpy(name, argstr, 16);
             break;
+#ifdef PR_PAC_RESET_KEYS
+        case(PR_PAC_RESET_KEYS):
+            if(arg & ~(PR_PAC_APIAKEY | PR_PAC_APIBKEY | PR_PAC_APDAKEY | PR_PAC_APDBKEY | PR_PAC_APGAKEY)) {
+                PyErr_SetString(PyExc_ValueError, "Unknown pointer authentication key");
+                return NULL;
+            }
+            break;
+#endif
         case(PR_SET_PDEATHSIG):
             if(arg < 0 || arg > SIGRTMAX) {
                 PyErr_SetString(PyExc_ValueError, "Unknown signal");
@@ -154,6 +174,30 @@ prctl_prctl(PyObject *self, PyObject *args)
                 return NULL;
             }
             break;
+#endif
+#ifdef PR_GET_SPECULATION_CTRL
+        case(PR_SET_SPECULATION_CTRL):
+            if(arg != PR_SPEC_STORE_BYPASS && arg != PR_SPEC_INDIRECT_BRANCH) {
+                PyErr_SetString(PyExc_ValueError, "Invalid speculation control setting");
+                return NULL;
+            }
+            if(arg2 != PR_SPEC_ENABLE
+               && arg2 != PR_SPEC_DISABLE
+               && arg2 != PR_SPEC_FORCE_DISABLE
+#ifdef PR_SPEC_DISABLE_NOEXEC
+               && arg2 != PR_SPEC_DISABLE_NOEXEC
+#endif
+            ) {
+                PyErr_SetString(PyExc_ValueError, "Invalid speculation control value");
+                return NULL;
+            }
+            /* Intentionally not breaking */
+        case(PR_GET_SPECULATION_CTRL):
+            if(arg != PR_SPEC_STORE_BYPASS && arg != PR_SPEC_INDIRECT_BRANCH) {
+                PyErr_SetString(PyExc_ValueError, "Invalid speculation control setting");
+                return NULL;
+            }
+            break;
 #endif
         case(PR_SET_TIMING):
             if(arg != PR_TIMING_STATISTICAL && arg != PR_TIMING_TIMESTAMP) {
@@ -199,14 +243,29 @@ prctl_prctl(PyObject *self, PyObject *args)
         case(PR_SET_ENDIAN):
         case(PR_SET_FPEMU):
         case(PR_SET_FPEXC):
+#ifdef PR_SET_FP_MODE
+        case(PR_GET_FP_MODE):
+        case(PR_SET_FP_MODE):
+#endif
+#ifdef PR_SET_IO_FLUSHER
+        case(PR_GET_IO_FLUSHER):
+        case(PR_SET_IO_FLUSHER):
+#endif
         case(PR_SET_KEEPCAPS):
         case(PR_GET_KEEPCAPS):
 #ifdef PR_MCE_KILL
         case(PR_GET_MCE_KILL):
 #endif
+#ifdef PR_MPX_ENABLE_MANAGEMENT
+        case(PR_MPX_ENABLE_MANAGEMENT):
+        case(PR_MPX_DISABLE_MANAGEMENT):
+#endif
 #ifdef PR_GET_NO_NEW_PRIVS
         case(PR_GET_NO_NEW_PRIVS):
         case(PR_SET_NO_NEW_PRIVS):
+#endif
+#ifdef PR_PAC_RESET_KEYS
+        case(PR_PAC_RESET_KEYS):
 #endif
         case(PR_SET_PDEATHSIG):
 #if defined(PR_GET_PTRACER) && (PR_GET_PTRACER != NOT_SET)
@@ -223,6 +282,18 @@ prctl_prctl(PyObject *self, PyObject *args)
         case(PR_SET_SECUREBITS):
         case(PR_GET_SECUREBITS):
 #endif
+#ifdef PR_GET_SPECULATION_CTRL
+        case(PR_GET_SPECULATION_CTRL):
+        case(PR_SET_SPECULATION_CTRL):
+#endif
+#ifdef PR_TASK_PERF_EVENTS_DISABLE
+        case(PR_TASK_PERF_EVENTS_DISABLE):
+        case(PR_TASK_PERF_EVENTS_ENABLE):
+#endif
+#ifdef PR_SET_THP_DISABLE
+        case(PR_GET_THP_DISABLE):
+        case(PR_SET_THP_DISABLE):
+#endif
 #ifdef PR_GET_TIMERSLACK
         case(PR_GET_TIMERSLACK):
         case(PR_SET_TIMERSLACK):
@@ -233,7 +304,7 @@ prctl_prctl(PyObject *self, PyObject *args)
         case(PR_SET_TSC):
 #endif
         case(PR_SET_UNALIGN):
-            result = prctl(option, arg, 0, 0, 0);
+            result = prctl(option, arg, arg2, 0, 0);
             if(result < 0) {
                 PyErr_SetFromErrno(PyExc_OSError);
                 return NULL;
@@ -243,12 +314,22 @@ prctl_prctl(PyObject *self, PyObject *args)
                 case(PR_CAPBSET_READ):
 #endif
                 case(PR_GET_DUMPABLE):
+#ifdef PR_SET_IO_FLUSHER
+                case(PR_GET_IO_FLUSHER):
+#endif
                 case(PR_GET_KEEPCAPS):
 #ifdef PR_GET_SECCOMP
                 case(PR_GET_SECCOMP):
+#endif
+#ifdef PR_GET_THP_DISABLE
+                case(PR_GET_THP_DISABLE):
 #endif
                 case(PR_GET_TIMING):
                     return PyBool_FromLong(result);
+
+#ifdef PR_SET_FP_MODE
+                case(PR_GET_FP_MODE):
+#endif
 #ifdef PR_MCE_KILL
                 case(PR_GET_MCE_KILL):
 #endif
@@ -261,6 +342,9 @@ prctl_prctl(PyObject *self, PyObject *args)
 #ifdef PR_GET_SECUREBITS
                 case(PR_GET_SECUREBITS):
 #endif
+#ifdef PR_GET_SPECULATION_CTRL
+                case(PR_GET_SPECULATION_CTRL):
+#endif
 #ifdef PR_GET_TIMERSLACK
                 case(PR_GET_TIMERSLACK):
 #endif
@@ -327,9 +411,6 @@ prctl_prctl(PyObject *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
-/* While not part of prctl, this complements PR_SET_NAME */
-static int __real_argc = -1;
-static char **__real_argv = NULL;
 #if PY_MAJOR_VERSION < 3
 #define _Py_GetArgcArgv Py_GetArgcArgv
 #else
@@ -366,6 +447,9 @@ static int _Py_GetArgcArgv(int* argc, char ***argv) {
     char **buf = NULL , *arg0 = NULL, *ptr = 0, *limit = NULL;
 
     Py_GetArgcArgv(argc, &argv_w);
+    if (*argc < 1 || argv_w == NULL) {
+        return 0;
+    }
 
     buf = (char **)malloc((*argc + 1) * sizeof(char *));
     buf[*argc] = NULL;
@@ -405,28 +489,24 @@ static PyObject *
 prctl_set_proctitle(PyObject *self, PyObject *args)
 {
     int argc = 0;
-    char **argv;
-    int len;
     char *title;
+    static size_t len = 0;
+    static char **argv;
+
     if(!PyArg_ParseTuple(args, "s", &title)) {
         return NULL;
     }
-    if(__real_argc > 0)  {
-        argc = __real_argc;
-        argv = __real_argv;
-    }
-    else {
+    if(argv == NULL)  {
         _Py_GetArgcArgv(&argc, &argv);
-        __real_argc = argc;
-        __real_argv = argv;
+        if(argc > 0) {
+            len = (size_t)(argv[argc-1]) + strlen(argv[argc-1]) - (size_t)(argv[0]);
+        }
     }
 
-    if(argc <= 0) {
+    if(len <= 0) {
         PyErr_SetString(PyExc_RuntimeError, "Failed to locate argc/argv");
         return NULL;
     }
-    /* Determine up to where we can write */
-    len = (size_t)(argv[argc-1]) + strlen(argv[argc-1]) - (size_t)(argv[0]);
     strncpy(argv[0], title, len);
     if(strlen(title) < len)
         memset(argv[0] + strlen(title), 0, len - strlen(title));
@@ -643,6 +723,9 @@ PyInit__prctl(void)
 #ifdef PR_CAPBSET_READ
     namedconstant(PR_CAPBSET_READ);
     namedconstant(PR_CAPBSET_DROP);
+#endif
+#ifdef PR_SET_CHILD_SUBREAPER
+    namedattribute(CHILD_SUBREAPER);
 #endif
     namedattribute(DUMPABLE);
     namedattribute(ENDIAN);
@@ -663,14 +746,37 @@ PyInit__prctl(void)
     namedconstant(PR_FP_EXC_NONRECOV);
     namedconstant(PR_FP_EXC_ASYNC);
     namedconstant(PR_FP_EXC_PRECISE);
+#ifdef PR_SET_FP_MODE
+    namedattribute(FP_MODE);
+    namedconstant(PR_FP_MODE_FR);
+    namedconstant(PR_FP_MODE_FRE);
+#endif
+#ifdef PR_SET_IO_FLUSHER
+    namedattribute(IO_FLUSHER);
+#endif
     namedattribute(KEEPCAPS);
 #ifdef PR_MCE_KILL
     namedattribute(MCE_KILL);
     namedconstant(PR_MCE_KILL_DEFAULT);
     namedconstant(PR_MCE_KILL_EARLY);
     namedconstant(PR_MCE_KILL_LATE);
+#endif
+#ifdef PR_MPX_ENABLE_MANAGEMENT
+    namedconstant(PR_MPX_ENABLE_MANAGEMENT);
+    namedconstant(PR_MPX_DISABLE_MANAGEMENT);
 #endif
     namedattribute(NAME);
+#ifdef PR_SET_NO_NEW_PRIVS
+    namedattribute(NO_NEW_PRIVS);
+#endif
+#ifdef PR_PAC_RESET_KEYS
+    namedconstant(PR_PAC_RESET_KEYS);
+    namedconstant(PR_PAC_APIAKEY);
+    namedconstant(PR_PAC_APIBKEY);
+    namedconstant(PR_PAC_APDAKEY);
+    namedconstant(PR_PAC_APDBKEY);
+    namedconstant(PR_PAC_APGAKEY);
+#endif
     namedattribute(PDEATHSIG);
 #ifdef PR_SET_PTRACER
     namedattribute(PTRACER);
@@ -678,19 +784,34 @@ PyInit__prctl(void)
 #ifdef PR_SET_PTRACER_ANY
     namedconstant(PR_SET_PTRACER_ANY);
 #endif
-#ifdef PR_SET_CHILD_SUBREAPER
-    namedattribute(CHILD_SUBREAPER);
-#endif
-#ifdef PR_SET_NO_NEW_PRIVS
-    namedattribute(NO_NEW_PRIVS);
-#endif
-
 #ifdef PR_GET_SECCOMP
     namedattribute(SECCOMP);
 #endif
 #ifdef PR_GET_SECUREBITS
     namedattribute(SECUREBITS);
 #endif
+#ifdef PR_SET_SPECULATION_CTRL
+    namedattribute(SPECULATION_CTRL);
+    namedconstant(PR_SPEC_STORE_BYPASS);
+    namedconstant(PR_SPEC_INDIRECT_BRANCH);
+    namedconstant(PR_SPEC_PRCTL);
+    namedconstant(PR_SPEC_ENABLE);
+    namedconstant(PR_SPEC_DISABLE);
+    namedconstant(PR_SPEC_FORCE_DISABLE);
+#ifdef PR_SPEC_DISABLE_NOEXEC
+    namedconstant(PR_SPEC_DISABLE_NOEXEC);
+#endif
+#endif
+#ifdef PR_TASK_PERF_EVENTS_DISABLE
+    namedconstant(PR_TASK_PERF_EVENTS_DISABLE);
+    namedconstant(PR_TASK_PERF_EVENTS_ENABLE);
+#endif
+#ifdef PR_SET_THP_DISABLE
+    namedattribute(THP_DISABLE);
+#endif
+#ifdef PR_GET_TID_ADDRESS
+    namedconstant(PR_GET_TID_ADDRESS);
+#endif
 #ifdef PR_GET_TIMERSLACK
     namedattribute(TIMERSLACK);
 #endif
@@ -705,59 +826,70 @@ PyInit__prctl(void)
     namedattribute(UNALIGN);
     namedconstant(PR_UNALIGN_NOPRINT);
     namedconstant(PR_UNALIGN_SIGBUS);
-#ifdef PR_GET_TID_ADDRESS
-    namedconstant(PR_GET_TID_ADDRESS);
-#endif
+
     /* Add the CAP_* constants too */
     namedconstant(CAP_EFFECTIVE);
     namedconstant(CAP_PERMITTED);
     namedconstant(CAP_INHERITABLE);
+    namedconstant(CAP_AUDIT_CONTROL);
+#ifdef CAP_AUDIT_READ
+    namedconstant(CAP_AUDIT_READ);
+#endif
+    namedconstant(CAP_AUDIT_WRITE);
+#ifdef CAP_BLOCK_SUSPEND
+    namedconstant(CAP_BLOCK_SUSPEND);
+#endif
+#ifdef CAP_BPF
+    namedconstant(CAP_BPF);
+#endif
     namedconstant(CAP_CHOWN);
     namedconstant(CAP_DAC_OVERRIDE);
     namedconstant(CAP_DAC_READ_SEARCH);
     namedconstant(CAP_FOWNER);
     namedconstant(CAP_FSETID);
-    namedconstant(CAP_KILL);
-    namedconstant(CAP_SETGID);
-    namedconstant(CAP_SETUID);
-    namedconstant(CAP_SETPCAP);
-    namedconstant(CAP_LINUX_IMMUTABLE);
-    namedconstant(CAP_NET_BIND_SERVICE);
-    namedconstant(CAP_NET_BROADCAST);
-    namedconstant(CAP_NET_ADMIN);
-    namedconstant(CAP_NET_RAW);
     namedconstant(CAP_IPC_LOCK);
     namedconstant(CAP_IPC_OWNER);
-    namedconstant(CAP_SYS_MODULE);
-    namedconstant(CAP_SYS_RAWIO);
-    namedconstant(CAP_SYS_CHROOT);
-    namedconstant(CAP_SYS_PTRACE);
-    namedconstant(CAP_SYS_PACCT);
-    namedconstant(CAP_SYS_ADMIN);
-    namedconstant(CAP_SYS_BOOT);
-    namedconstant(CAP_SYS_NICE);
-    namedconstant(CAP_SYS_RESOURCE);
-    namedconstant(CAP_SYS_TIME);
-    namedconstant(CAP_SYS_TTY_CONFIG);
-    namedconstant(CAP_MKNOD);
+    namedconstant(CAP_KILL);
     namedconstant(CAP_LEASE);
-    namedconstant(CAP_AUDIT_WRITE);
-    namedconstant(CAP_AUDIT_CONTROL);
-#ifdef CAP_SETFCAP
-    namedconstant(CAP_SETFCAP);
-#endif
+    namedconstant(CAP_LINUX_IMMUTABLE);
 #ifdef CAP_MAC_OVERRIDE
     namedconstant(CAP_MAC_OVERRIDE);
 #endif
 #ifdef CAP_MAC_ADMIN
     namedconstant(CAP_MAC_ADMIN);
 #endif
+    namedconstant(CAP_MKNOD);
+    namedconstant(CAP_NET_ADMIN);
+    namedconstant(CAP_NET_BIND_SERVICE);
+    namedconstant(CAP_NET_BROADCAST);
+    namedconstant(CAP_NET_RAW);
+#ifdef CAP_PERFMON
+    namedconstant(CAP_PERFMON);
+#endif
+#ifdef CAP_SETFCAP
+    namedconstant(CAP_SETFCAP);
+#endif
+    namedconstant(CAP_SETGID);
+    namedconstant(CAP_SETPCAP);
+    namedconstant(CAP_SETUID);
+    namedconstant(CAP_SYS_ADMIN);
+    namedconstant(CAP_SYS_BOOT);
+    namedconstant(CAP_SYS_CHROOT);
+    namedconstant(CAP_SYS_MODULE);
+    namedconstant(CAP_SYS_NICE);
+    namedconstant(CAP_SYS_PACCT);
+    namedconstant(CAP_SYS_PTRACE);
+    namedconstant(CAP_SYS_RAWIO);
+    namedconstant(CAP_SYS_RESOURCE);
+    namedconstant(CAP_SYS_TIME);
+    namedconstant(CAP_SYS_TTY_CONFIG);
 #ifdef CAP_SYSLOG
     namedconstant(CAP_SYSLOG);
 #endif
 #ifdef CAP_WAKE_ALARM
     namedconstant(CAP_WAKE_ALARM);
 #endif
+    namedconstant(CAP_LAST_CAP);
     /* And the securebits constants */
     namedconstant(SECURE_KEEP_CAPS);
     namedconstant(SECURE_NO_SETUID_FIXUP);
